"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const hatchet_client_1 = require("@clients/hatchet-client");
const dispatcher_1 = require("@protoc/dispatcher");
const action_listener_1 = require("@clients/dispatcher/action-listener");
const nice_grpc_common_1 = require("nice-grpc-common");
const action_listener_test_1 = require("@clients/dispatcher/action-listener.test");
const zod_1 = require("zod");
const sleep_1 = __importDefault(require("@util/sleep"));
const nice_grpc_1 = require("nice-grpc");
const worker_1 = require("./worker");
const mockStart = {
    tenantId: 'TENANT_ID',
    jobId: 'job1',
    jobName: 'Job One',
    jobRunId: 'run1',
    stepId: 'step1',
    stepRunId: 'runStep1',
    actionId: 'action1',
    actionType: dispatcher_1.ActionType.START_STEP_RUN,
    actionPayload: JSON.stringify('{"input": {"data": 1}}'),
};
const mockCancel = Object.assign(Object.assign({}, mockStart), { actionType: dispatcher_1.ActionType.CANCEL_STEP_RUN });
describe('Worker', () => {
    let hatchet;
    beforeEach(() => {
        hatchet = new hatchet_client_1.HatchetClient({
            token: 'TOKEN',
            log_level: 'OFF',
            host_port: 'HOST_PORT',
            tls_config: {
                cert_file: 'TLS_CERT_FILE',
                key_file: 'TLS_KEY_FILE',
                ca_file: 'TLS_ROOT_CA_FILE',
                server_name: 'TLS_SERVER_NAME',
            },
        }, {
            credentials: nice_grpc_1.ChannelCredentials.createInsecure(),
        });
    });
    describe('register_workflow', () => {
        it('should update the registry', () => __awaiter(void 0, void 0, void 0, function* () {
            const worker = new worker_1.Worker(hatchet, { name: 'WORKER_NAME' });
            const putWorkflowSpy = jest.spyOn(worker.client.admin, 'put_workflow').mockResolvedValue();
            const workflow = {
                id: 'workflow1',
                description: 'test',
                on: {
                    event: 'user:create',
                },
                steps: [
                    {
                        name: 'step1',
                        run: (ctx) => {
                            return { test: 'test' };
                        },
                    },
                ],
            };
            yield worker.register_workflow(workflow);
            expect(putWorkflowSpy).toHaveBeenCalledTimes(1);
            expect(worker.action_registry).toEqual({
                [`default:step1`]: workflow.steps[0].run,
            });
        }));
    });
    describe('handle_start_step_run', () => {
        it('should start a step run', () => __awaiter(void 0, void 0, void 0, function* () {
            const worker = new worker_1.Worker(hatchet, { name: 'WORKER_NAME' });
            const putWorkflowSpy = jest.spyOn(worker.client.admin, 'put_workflow').mockResolvedValue();
            const getActionEventSpy = jest.spyOn(worker, 'get_action_event');
            const sendActionEventSpy = jest
                .spyOn(worker.client.dispatcher, 'send_action_event')
                .mockResolvedValue({
                tenantId: 'TENANT_ID',
                workerId: 'WORKER_ID',
            });
            const startSpy = jest.fn().mockReturnValue({ data: 4 });
            worker.action_registry = {
                [mockStart.actionId]: startSpy,
            };
            worker.handle_start_step_run(mockStart);
            yield (0, sleep_1.default)(100);
            expect(startSpy).toHaveBeenCalledTimes(1);
            expect(getActionEventSpy).toHaveBeenNthCalledWith(2, expect.anything(), dispatcher_1.ActionEventType.STEP_EVENT_TYPE_COMPLETED, { data: 4 });
            expect(worker.futures[mockStart.stepRunId]).toBeUndefined();
            expect(sendActionEventSpy).toHaveBeenCalledTimes(2);
        }));
        it('should fail gracefully', () => __awaiter(void 0, void 0, void 0, function* () {
            const worker = new worker_1.Worker(hatchet, { name: 'WORKER_NAME' });
            const getActionEventSpy = jest.spyOn(worker, 'get_action_event');
            const sendActionEventSpy = jest
                .spyOn(worker.client.dispatcher, 'send_action_event')
                .mockResolvedValue({
                tenantId: 'TENANT_ID',
                workerId: 'WORKER_ID',
            });
            const startSpy = jest.fn().mockRejectedValue(new Error('ERROR'));
            worker.action_registry = {
                [mockStart.actionId]: startSpy,
            };
            worker.handle_start_step_run(mockStart);
            yield (0, sleep_1.default)(100);
            expect(startSpy).toHaveBeenCalledTimes(1);
            expect(getActionEventSpy).toHaveBeenNthCalledWith(2, expect.anything(), dispatcher_1.ActionEventType.STEP_EVENT_TYPE_FAILED, expect.anything());
            expect(worker.futures[mockStart.stepRunId]).toBeUndefined();
            expect(sendActionEventSpy).toHaveBeenCalledTimes(2);
        }));
    });
    describe('handle_cancel_step_run', () => {
        it('should cancel a step run', () => {
            const worker = new worker_1.Worker(hatchet, { name: 'WORKER_NAME' });
            const cancelSpy = jest.fn().mockReturnValue(undefined);
            worker.futures = {
                [mockCancel.stepRunId]: {
                    cancel: cancelSpy,
                },
            };
            worker.handle_cancel_step_run(mockCancel);
            expect(cancelSpy).toHaveBeenCalledTimes(1);
            expect(worker.futures[mockCancel.stepRunId]).toBeUndefined();
        });
    });
    describe('exit_gracefully', () => {
        xit('should call exit_gracefully on SIGTERM', () => __awaiter(void 0, void 0, void 0, function* () {
            const worker = new worker_1.Worker(hatchet, { name: 'WORKER_NAME' });
            // the spy is not working and the test is killing the test process
            const exitSpy = jest.spyOn(worker, 'exit_gracefully').mockImplementationOnce(() => {
                throw new Error('Simulated error');
            });
            process.emit('SIGTERM', 'SIGTERM');
            expect(exitSpy).toHaveBeenCalledTimes(1);
        }));
        xit('should call exit_gracefully on SIGINT', () => __awaiter(void 0, void 0, void 0, function* () {
            const worker = new worker_1.Worker(hatchet, { name: 'WORKER_NAME' });
            // This is killing the process (as it should) fix the spy at some point
            const exitSpy = jest.spyOn(worker, 'exit_gracefully').mockResolvedValue();
            process.emit('SIGINT', 'SIGINT');
            expect(exitSpy).toHaveBeenCalledTimes(1);
        }));
        xit('should unregister the listener and exit', () => __awaiter(void 0, void 0, void 0, function* () {
            const worker = new worker_1.Worker(hatchet, { name: 'WORKER_NAME' });
            jest.spyOn(process, 'exit').mockImplementation((number) => {
                throw new Error(`EXIT ${number}`);
            }); // This is killing the process (as it should) fix the spy at some point
            const mockActionListener = new action_listener_1.ActionListener(hatchet.dispatcher, (0, action_listener_test_1.mockListener)([mockStart, mockStart, new nice_grpc_common_1.ServerError(nice_grpc_common_1.Status.CANCELLED, 'CANCELLED')]), 'WORKER_ID');
            mockActionListener.unregister = jest.fn().mockResolvedValue((0, zod_1.never)());
            worker.listener = mockActionListener;
            expect(() => __awaiter(void 0, void 0, void 0, function* () {
                yield worker.exit_gracefully();
            })).toThrow('EXIT 0');
            expect(mockActionListener.unregister).toHaveBeenCalledTimes(1);
        }));
        it('should exit the process if handle_kill is true', () => __awaiter(void 0, void 0, void 0, function* () {
            const worker = new worker_1.Worker(hatchet, { name: 'WORKER_NAME' });
            const exitSpy = jest.spyOn(process, 'exit').mockReturnValue(undefined);
            yield worker.exit_gracefully();
            expect(exitSpy).toHaveBeenCalledTimes(1);
        }));
    });
    describe('start', () => {
        xit('should get actions and start runs', () => __awaiter(void 0, void 0, void 0, function* () {
            const worker = new worker_1.Worker(hatchet, { name: 'WORKER_NAME' });
            const startSpy = jest.spyOn(worker, 'handle_start_step_run').mockReturnValue();
            const cancelSpy = jest.spyOn(worker, 'handle_cancel_step_run').mockReturnValue();
            const mockActionListener = new action_listener_1.ActionListener(hatchet.dispatcher, (0, action_listener_test_1.mockListener)([mockStart, mockStart, new nice_grpc_common_1.ServerError(nice_grpc_common_1.Status.CANCELLED, 'CANCELLED')]), 'WORKER_ID');
            const getActionListenerSpy = jest
                .spyOn(worker.client.dispatcher, 'get_action_listener')
                .mockResolvedValue(mockActionListener);
            yield worker.start();
            expect(getActionListenerSpy).toHaveBeenCalledTimes(1);
            expect(startSpy).toHaveBeenCalledTimes(2);
            expect(cancelSpy).toHaveBeenCalledTimes(0);
        }));
        it('should get actions and cancel runs', () => __awaiter(void 0, void 0, void 0, function* () {
            const worker = new worker_1.Worker(hatchet, { name: 'WORKER_NAME' });
            const startSpy = jest.spyOn(worker, 'handle_start_step_run').mockReturnValue();
            const cancelSpy = jest.spyOn(worker, 'handle_cancel_step_run').mockReturnValue();
            const mockActionListner = new action_listener_1.ActionListener(hatchet.dispatcher, (0, action_listener_test_1.mockListener)([mockStart, mockCancel, new nice_grpc_common_1.ServerError(nice_grpc_common_1.Status.CANCELLED, 'CANCELLED')]), 'WORKER_ID');
            const getActionListenerSpy = jest
                .spyOn(worker.client.dispatcher, 'get_action_listener')
                .mockResolvedValue(mockActionListner);
            yield worker.start();
            expect(getActionListenerSpy).toHaveBeenCalledTimes(1);
            expect(startSpy).toHaveBeenCalledTimes(1);
            expect(cancelSpy).toHaveBeenCalledTimes(1);
        }));
        it('should retry 5 times to start a worker then throw an error', () => __awaiter(void 0, void 0, void 0, function* () {
            const worker = new worker_1.Worker(hatchet, { name: 'WORKER_NAME' });
            const startSpy = jest.spyOn(worker, 'handle_start_step_run').mockReturnValue();
            const cancelSpy = jest.spyOn(worker, 'handle_cancel_step_run').mockReturnValue();
            const mockActionListner = new action_listener_1.ActionListener(hatchet.dispatcher, (0, action_listener_test_1.mockListener)([mockStart, mockCancel, new nice_grpc_common_1.ServerError(nice_grpc_common_1.Status.CANCELLED, 'CANCELLED')]), 'WORKER_ID');
            const getActionListenerSpy = jest
                .spyOn(worker.client.dispatcher, 'get_action_listener')
                .mockImplementationOnce(() => {
                throw new Error('Simulated error');
            })
                .mockImplementationOnce(() => {
                throw new Error('Simulated error');
            })
                .mockImplementationOnce(() => {
                throw new Error('Simulated error');
            })
                .mockImplementationOnce(() => {
                throw new Error('Simulated error');
            })
                .mockImplementationOnce(() => {
                throw new Error('Simulated error');
            })
                .mockImplementationOnce(() => {
                throw new Error('Simulated error');
            });
            yield worker.start();
            expect(getActionListenerSpy).toHaveBeenCalledTimes(5);
            expect(startSpy).toHaveBeenCalledTimes(0);
            expect(cancelSpy).toHaveBeenCalledTimes(0);
        }));
        it('should successfully run after retrying < 5 times', () => __awaiter(void 0, void 0, void 0, function* () {
            const worker = new worker_1.Worker(hatchet, { name: 'WORKER_NAME' });
            const startSpy = jest.spyOn(worker, 'handle_start_step_run').mockReturnValue();
            const cancelSpy = jest.spyOn(worker, 'handle_cancel_step_run').mockReturnValue();
            const mockActionLister = new action_listener_1.ActionListener(hatchet.dispatcher, (0, action_listener_test_1.mockListener)([mockStart, mockCancel, new nice_grpc_common_1.ServerError(nice_grpc_common_1.Status.CANCELLED, 'CANCELLED')]), 'WORKER_ID');
            const getActionListenerSpy = jest
                .spyOn(worker.client.dispatcher, 'get_action_listener')
                .mockImplementationOnce(() => {
                throw new Error('Simulated error');
            })
                .mockImplementationOnce(() => {
                throw new Error('Simulated error');
            })
                .mockImplementationOnce(() => {
                throw new Error('Simulated error');
            })
                .mockImplementationOnce(() => {
                throw new Error('Simulated error');
            })
                .mockImplementationOnce(() => __awaiter(void 0, void 0, void 0, function* () { return mockActionLister; }));
            yield worker.start();
            expect(getActionListenerSpy).toHaveBeenCalledTimes(5);
            expect(startSpy).toHaveBeenCalledTimes(1);
            expect(cancelSpy).toHaveBeenCalledTimes(1);
        }));
    });
});
